# https://pandas.pydata.org/pandas-docs/stable/user_guide/10min.html
import numpy as np
import pandas as pd

s = pd.Series([1, 3, 5, np.nan, 6, 8])
dates = pd.date_range("20130101", periods=6)
data = np.array([[ 1.14, -1.23,  0.4 , -0.68],
       [-0.87, -0.58, -0.31,  0.06],
       [-1.17,  0.9 ,  0.47, -1.54],
       [ 1.49,  1.9 ,  1.18, -0.18],
       [-1.07,  1.05, -0.4 ,  1.22],
       [ 0.21,  0.98,  0.36,  0.71]])
df = pd.DataFrame(data, index=dates, columns=list("ABCD"))

df.loc["20130103"] # <-- fails because "20130103" is not an exact string match for date string in JSON



import babypandas as bpd

phones = bpd.DataFrame().assign(
    Model=['iPhone 13', 'iPhone 13 Pro Max', 'Samsung Galaxy Z Flip', 'Pixel 5a'],
    Price=[799, 1099, 999, 449],
    Screen=[6.1, 6.7, 6.7, 6.3]
)

inventory = bpd.DataFrame().assign(
    Handset=['iPhone 13 Pro Max', 'iPhone 13', 'Pixel 5a', 'iPhone 13'],
    Units=[50, 40, 10, 100],
    Store=['Westfield UTC', 'Westfield UTC', 'Fashion Valley', 'Downtown']
)

phones.merge(inventory, left_on='Model', right_on='Handset')




# https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html
import pandas as pd
import numpy as np

df3 = pd.DataFrame({"X": ["A", "B", "A", "B"], "Y": [1, 4, 3, 2]})
df3.groupby(["X"]).get_group("A")


# tests that pointers still trigger memory cap


class MyStr:
    def __init__(self, val):
        self.val = val
    def __repr__(self):
        return f'"{self.val}"'


# this uses about 130 MB
# but, sys.getsizeof() only returns ~120 KB (updated 5/13/2024, don't know how big sys.getsizeof() thinks it is)
hello = [MyStr("hello world") for _ in range(1_000_000)]

hello[:5]



import pandas as pd
import io

csv = '''
breed,type,longevity,size,weight
German Shepherd,herding,9.73,large,
Beagle,hound,12.3,small,
Yorkshire Terrier,toy,12.6,small,5.5
Golden Retriever,sporting,12.04,medium,60.0
Bulldog,non-sporting,6.29,medium,45.0
Labrador Retriever,sporting,12.04,medium,67.5
Boxer,working,8.81,medium,
Poodle,non-sporting,11.95,medium,
Dachshund,hound,12.63,small,24.0
Rottweiler,working,9.11,large,
Boston Terrier,non-sporting,10.92,medium,
Shih Tzu,toy,13.2,small,12.5
Miniature Schnauzer,terrier,11.81,small,15.5
Doberman Pinscher,working,10.33,large,
Chihuahua,toy,16.5,small,5.5
Siberian Husky,working,12.58,medium,47.5
Pomeranian,toy,9.67,small,5.0
French Bulldog,non-sporting,9.0,medium,27.0
Great Dane,working,6.96,large,
Shetland Sheepdog,herding,12.53,small,22.0
Cavalier King Charles Spaniel,toy,11.29,small,15.5
German Shorthaired Pointer,sporting,11.46,large,62.5
Maltese,toy,12.25,small,5.0
'''

dogs = pd.read_csv(io.StringIO(csv))

(dogs
 .groupby('type').plot() # <-- crashes when i try to do a plot() here
)



# do not run this in the normal test suite!
# was just trying to give pg a very large JSON for his backend

import pandas as pd
from pathlib import Path

baby = Path('datasets/babynames.csv').resolve()

df = pd.read_csv(baby)

df.head(5)



import pandas as pd
import io

csv = """
breed,type,longevity,size,temp
Labrador,sporting,12.04,medium,1
German,herding,9.73,large,2
Beagle,hound,12.3,small,3
Golden,sporting,12.04,medium,4
Yorkshire,toy,12.6,small,5
Bulldog,non-sporting,6.29,medium,6
Boxer,working,8.81,medium,7
Poodle,non-sporting,11.95,medium,8
"""

dogs = pd.read_csv(io.StringIO(csv))

dogs.groupby("size")[["longevity", "temp"]].transform(lambda s: s.mean())



import io

import pandas as pd

csv = """
breed,type,longevity,size
Labrador,sporting,12.04,medium
German,herding,9.73,large
Beagle,hound,12.3,small
Golden,sporting,12.04,medium
Yorkshire,toy,12.6,small
Bulldog,non-sporting,6.29,medium
Boxer,working,8.81,medium
Poodle,non-sporting,11.95,medium
"""

dogs = pd.read_csv(io.StringIO(csv))

dogs.groupby("size")[['longevity']].transform(lambda s: s.mean())



import pandas as pd
import io

csv = """
Player,Position,Team,Salary
Stephen Curry,PG,Golden State Warriors,45780966
Russell Westbrook,PG,Los Angeles Lakers,44211146
James Harden,PG,Brooklyn Nets,43848000
LeBron James,SF,Los Angeles Lakers,41180544
Kevin Durant,PF,Brooklyn Nets,40918900
"""

df = pd.read_csv(io.StringIO(csv)).set_index("Player")

series = df["Salary"]
series["Kevin Durant"]



# these will all get parsed into PassThroughCalls since we can only robustly
# create GroupbyFilerCalls during runtime analysis, so this test isn't really
# testing that much, but oh well!

(dogs
 .groupby('size')
 .filter(lambda df: df.shape[0] > 2)
)

(dogs
 .groupby('size')
 .filter(lambda df: df.shape[0] > 2, dropna=False)
)

# no groupby, so it shouldn't get parsed into a GroupByFilterCall
df.filter(items=['one', 'three'])

# can't handle this at parse time, so we'll just pass it through
(dogs
 .groupby('size')
 ['weight']
 .filter(lambda x: x.shape[0] > 2, dropna=False)
)


(dogs
 .groupby('grooming')
 .agg('mean')
)

(dogs
 .groupby('grooming')
 .var()
)

# Note that this is an agg but we have to handle it at runtime, not parse time
(dogs
 .groupby('grooming')
 ['weight']
 .var()
)

# should not be parsed into an AggCall
(dogs
 .groupby('grooming')
 .transform(lambda x: x.max() - x.min())
)



# https://github.com/SamLau95/pandas_tutor/issues/32
# https://pandas.pydata.org/pandas-docs/stable/user_guide/10min.html
import numpy as np
import pandas as pd

dates = pd.date_range("20130101", periods=6)
data = np.array([[ 1.14, -1.23,  0.4 , -0.68],
       [-0.87, -0.58, -0.31,  0.06],
       [-1.17,  0.9 ,  0.47, -1.54],
       [ 1.49,  1.9 ,  1.18, -0.18],
       [-1.07,  1.05, -0.4 ,  1.22],
       [ 0.21,  0.98,  0.36,  0.71]])
df = pd.DataFrame(data, index=dates, columns=list("ABCD"))
df2 = df.copy()
df2["E"] = ["one", "one", "two", "three", "four", "three"]

df2[df2["E"].isin(["two", "four"])] # <-- is this an example of a nested expression?



"""
defines %%pandas_tutor magic
"""
import ast
import json

from IPython.core.magic import (
    Magics,
    line_cell_magic,
    cell_magic,
    magics_class,
)
from IPython.display import HTML, display

import pandas_tutor.__main__ as pt
import pandas_tutor.util as util

if util.in_dev():
    wsembed_bundle_url = "https://cokapi.com/wst-pg-devel/wst/wsapp/frontend/build/wsembed.bundle.js"  # noqa: E501
else:
    wsembed_bundle_url = "https://pandastutor.com/build/wsembed.bundle.2022-07-07-release.js"  # noqa: E501

# when using projectors we need stronger colors, otherwise people can't
# distinguish between colors
#
# colors from https://colorbrewer2.org/#type=qualitative&scheme=Set1&n=9
projector_colors = [
    "#e41a1c",
    "#377eb8",
    "#4daf4a",
    "#984ea3",
    "#ff7f00",
    "#a65628",
    "#f781bf",
    "#999999",
]

default_wsv_colors = [
    "#a6cee3",
    "#b2df8a",
    "#fb9a99",
    "#fdbf6f",
    "#cab2d6",
    "#ff7f00",
    "#33a02c",
    "#6a3d9a",
    "#e31a1c",
    "#1f78b4",
    "#b15928",
    "#ffff99",
]


# runs on initial load to load the wst js library
#
# HACK: use setTimeout to avoid race condition with loading pandas tutor JS
_load_wst_bundle = f"""
<script type="text/javascript" src="{wsembed_bundle_url}"></script>
<script>
console.log("initializing pandas_tutor js")

function drawWsv(viz_id, spec, options) {{
  if (typeof createWsvFromPandasTrace === 'undefined') {{
    setTimeout(() => drawWsv(viz_id, spec, options), 2000) // retry in 2 seconds
    return
  }}
  createWsvFromPandasTrace(viz_id, spec, options)
}}
</script>
"""  # noqa: E501

# displays each time a cell with %%pandas_tutor is run
_viz_html = """
<div class="pt-viz" id="{viz_id}"></div>
<script>
drawWsv('#{viz_id}', {spec}, {options});
</script>
"""


@magics_class
class PandasTutorMagics(Magics):

    viz_count = 0

    display_options: dict = {
        "nohover": True,
        "maxDisplayRows": 7,
        "maxDisplayCols": 5,
    }

    def __init__(self, shell, **kwargs):
        super().__init__(shell, **kwargs)
        display(HTML(_load_wst_bundle))

    @line_cell_magic
    def set_pandas_tutor_options(self, line: str, cell=None):
        """
        used to set display options using a python dict. example:

        %%set_pandas_tutor_options
        { "maxDisplayRows": 10, "maxDisplayCols": 5 }

        options:

        - maxDisplayRows: maximum number of rows to display in the table
        - maxDisplayCols: maximum number of columns to display in the table
        - nohover: if True, the table will not show hover effects
        - projectorMode: if True, use bolder colors for projectors
        """
        opts = ast.literal_eval(cell if cell else line)
        if not isinstance(opts, dict):
            raise ValueError("options must be a dict")

        if "projectorMode" in opts:
            if opts.pop("projectorMode"):
                self.display_options["colorPalette"] = projector_colors
            else:
                # TODO: can we just delete the key? neither way seems to
                # update the wsv colors after setting projectorMode
                self.display_options["colorPalette"] = default_wsv_colors
        self.display_options.update(opts)

    @cell_magic
    def pandas_tutor(self, line: str, cell: str):
        # inherits self.shell from Magics parent class
        spec = pt.make_tutor_spec_ipython(cell, self.shell)
        # need to double wrap so that quotes and \n are escaped
        spec_json = json.dumps(spec)

        viz_id = f"pt-viz-{self.viz_count}"
        options = json.dumps(self.display_options)
        make_viz = _viz_html.format(
            viz_id=viz_id, spec=spec_json, options=options
        )
        display(HTML(make_viz))

        # increment so that each cell gets a unique css id
        self.viz_count += 1

    # %%pt is an alias for %%pandas_tutor
    @cell_magic
    def pt(self, line: str, cell: str):
        return self.pandas_tutor(line, cell)



import pandas as pd

# when after index look like positional index, but it is actually group names.
# Use this test case to make sure we are not providing the wrong information
# TODO: rewrite the golden file once groupers and custom function groupby are
# implemented
data = {"Group": ["A", "B", "C", "D", "E"], "Value": [4, 3, 2, 1, 0]}

# Creating DataFrame
df = pd.DataFrame(data)


def single_group(value):
    return value + 1


df.groupby(single_group).apply(lambda x: x.iloc[0])



"""
runs pandas_tutor backend

Usage:
    pandas_tutor FILE ... [--output] [--parse_only] [--parse_log]
    pandas_tutor -c CODE

Options:
    -o --output      # Outputs specs to files named {input_file}.golden
    -p --parse_only  # Outputs parsed code rather than full spec
    -l --parse_log   # Outputs parse debug output
    -c --code        # Code as a string (instead of a file)
"""

import argparse
from pathlib import Path

from .diagram import OutputSpec
from .parse import parse, parse_as_json, test_logger
from .run import run
from .serialize import serialize


def make_tutor_spec(code: str) -> str:
    """oh yeah, it's all coming together"""
    root = parse(code)
    eval_results = run(root)
    explanation = serialize(eval_results)
    spec = OutputSpec(code=code, explanation=explanation)
    return spec.to_json()


def make_tutor_spec_ipython(code: str, ipython_shell) -> str:
    """
    when we run in ipython, we need to execute code using ipython's namespace
    """
    root = parse(code)
    eval_results = run(root, ipython_shell)
    explanation = serialize(eval_results)
    spec = OutputSpec(code=code, explanation=explanation)
    return spec.to_json()


def make_tutor_spec_py(code: str) -> OutputSpec:
    """Keeps serialized output as a Python object for testing"""
    root = parse(code)
    eval_results = run(root)
    explanation = serialize(eval_results)
    spec = OutputSpec(code=code, explanation=explanation)
    return spec


def spec_from_file(filename: str, spec_fn=make_tutor_spec) -> str:
    code = Path(filename).read_text()
    return spec_fn(code)


def write_spec_to_file(spec: str, out: Path) -> None:
    print(f"Writing {out}")
    with out.open("w") as f:
        f.write(spec)


def create_parser():
    """Create argument parser"""
    parser = argparse.ArgumentParser(
        description="runs pandas_tutor backend",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # Add positional arguments for files
    parser.add_argument(
        "files", nargs="*", metavar="FILE", help="Input files to process"
    )

    # Add optional arguments
    parser.add_argument(
        "-c",
        "--code",
        metavar="CODE",
        help="Code as a string (instead of a file)",
    )
    parser.add_argument(
        "-o",
        "--output",
        action="store_true",
        help="Outputs specs to files named {input_file}.golden",
    )
    parser.add_argument(
        "-p",
        "--parse_only",
        action="store_true",
        help="Outputs parsed code rather than full spec",
    )
    parser.add_argument(
        "-l",
        "--parse_log",
        action="store_true",
        help="Outputs parse debug output",
    )

    parser.add_argument("--version", action="version", version="1.0")

    return parser


def main():
    """Main entry point for the pandas_tutor command"""
    parser = create_parser()
    args = parser.parse_args()

    # Validate that either files or code is provided
    if not args.code and not args.files:
        parser.error("Must provide either files or --code argument")

    spec_fn = (
        test_logger
        if args.parse_log
        else parse_as_json
        if args.parse_only
        else make_tutor_spec
    )

    if args.code:
        print(spec_fn(args.code))

    if args.files:
        if not args.output:
            for filename in args.files:
                print(spec_from_file(filename, spec_fn))  # type: ignore
        else:
            for filename in args.files:
                spec = spec_from_file(filename, spec_fn)  # type: ignore
                out_filename = Path(filename + ".golden")
                write_spec_to_file(spec, out_filename)


if __name__ == "__main__":
    main()
